/*
* Copyright (C) 2006 Roberto Alsina <ralsina@kde.org>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <vauth.h>
#include <vpopmail.h>
#include "bstrlib.h"
#include <sys/types.h>
#include <unistd.h>


// This is defined in libvpopmail.a but it's not on the headers.
// If anyone knows the "right" way to get quota usage info, I
// would like to know it.

int vmaildir_readquota (const char *dir, const char *quota);

// Converts an envvar to an int
int
envtoint (char *vname, int defval)
{
  bstring envvar_;
  int retval = defval;

  envvar_ = bfromcstr (getenv (vname));
  if (envvar_)
    {
      retval = atoi (envvar_->data);
      bdestroy (envvar_);
    }
  //printf ("%s=%d\n",vname,retval);
  return retval;
}

bstring
envtostr (char *vname)
{
  bstring retval;
  retval = bfromcstr (getenv (vname));
  if (!retval)
    retval = bfromcstr ("");

  //printf ("%s=%s\n",vname,retval->data);
  return retval;
}

int
main ()
{
  struct vqpasswd *user_passwd = NULL;

  int wrongrcptcount = envtoint ("WRONGRCPTCOUNT", 0);
  int chkuser_mbxquota = envtoint ("CHKUSER_MBXQUOTA", 100);
  int chkuser_rcptlimit = envtoint ("CHKUSER_RCPTLIMIT", 10000);
  int chkuser_wrongrcptlimit = envtoint ("CHKUSER_WRONGRCPTLIMIT", 10000);
  int smtprcptcount = envtoint ("SMTPRCPTCOUNT", 0);
  int ppid = getppid ();

  //Reset wrongrcptcount on new mail 
  if (wrongrcptcount > smtprcptcount)
    wrongrcptcount = 0;

  bstring smtprcptto = envtostr ("SMTPRCPTTO");

  struct bstrList *pieces = bsplit (smtprcptto, '@');
  if (pieces->qty != 2)
    {
      printf ("E511 Invalid address (#5.1.1 - chkuser2)\n");
      fprintf (stderr, "rcptchecks: pid %d - 511 Invalid address (%s)\n",
               ppid, smtprcptto->data);
      exit (0);
    }
  bstring username = pieces->entry[0];
  bstring domain = pieces->entry[1];

  // If there are more than $CHKUSER_RCPTLIMIT RCPTs accepted
  // fail with error

  if (chkuser_rcptlimit <= smtprcptcount)
    {
      printf
        ("E571 sorry, reached maximum number of recipients for one session (#5.7.1 - chkuser2)\n");
      fprintf (stderr,
               "rcptchecks: pid %d - 571 sorry, reached maximum number of recipients for one session (%d of %d - %s)\n",
               ppid, smtprcptcount, chkuser_rcptlimit, smtprcptto->data);
      exit (0);
    }

  // If it is for one of our RCPTHOSTS, then check if the recipient
  // domain is in virtualdomains. We only check for domain.tld:domain.tld
  // because that is how vpopmail does it, and this plugin is for
  // vpopmail only

  bstring cfname = bfromcstr ("/var/qmail/control/virtualdomains");
  FILE *cfile = fopen (cfname->data, "r");

  if (!cfile)
    {

      // Well, maybe we have no virtual domains. No check possible.
      // Log anyway, because this plugin is usseless in this case!

      fprintf (stderr, "rcptchecks: pid %d - error opening %s\n", ppid,
               cfname->data);
      exit (0);
    }

  bstring vrcptdom = bformat ("%s:%s", domain->data, domain->data);

  int islocal = 0;
  for (;;)
    {
      bstring vdom = bgets ((bNgetc) fgetc, cfile, '\n');
      if (!vdom)                //End of stream, so this must not be a virtualdomain
        {
          islocal = 0;
          break;
        }

      //You may lack a \n at the last line :-P
      if (vdom->slen > 0 && vdom->data[vdom->slen - 1] == '\n')
        {
          btrunc (vdom, vdom->slen - 1);
        }

      if (0 == bstrcmp (vdom, vrcptdom))
        {
          //It's a local domain, so break and continue testing
          islocal = 1;
          break;
        }
    }

  // Since it is virtual here, check if the user exists.
  // IMPORTANT: Requires SPP 0.42 

  // We keep the number of failed RCPTs on WRONGRCPTCOUNT

  int valid = 1;
  if (islocal == 1)
    {
      user_passwd = vauth_getpw (username->data, domain->data);
      if (user_passwd == NULL)
        {
          //Maybe it's an alias
          bstring dest =
            bfromcstr (valias_select (username->data, domain->data));
          if (dest == NULL)
            {
              valid = 0;
            }
          else
            {
              pieces = bsplit (dest, '@');
              username = pieces->entry[0];
              domain = pieces->entry[1];
              user_passwd = vauth_getpw (username->data, domain->data);
              if (user_passwd == NULL)  //The alias points to nowhere
                {
                  printf
                    ("E511 sorry, no mailbox here by that name (#5.1.1 - chkuser2)\n");
                  fprintf (stderr,
                           "rcptchecks: pid %d - 511 sorry, no mailbox here by that name (%s@%s)\n",
                           ppid, username->data, domain->data);
                  exit (0);
                }
            }
        }
      // if valid=1, then he exists. So, we check if he is overquota
      if (valid == 1)
        {
          if (strcasecmp (user_passwd->pw_shell, "NOQUOTA") != 0)       //He has a quota
            {
              bstring path;
              bcatcstr (path = bfromcstr (user_passwd->pw_dir), "/Maildir");
              if (vmaildir_readquota
                  (path->data,
                   format_maildirquota (user_passwd->pw_shell)) >=
                  chkuser_mbxquota)
                {
                  printf
                    ("E522 sorry, recipient mailbox is full (#5.2.2 - chkuser2)\n");
                  fprintf (stderr,
                           "rcptchecks: pid %d - 522 sorry, recipient mailbox is full (%s@%s)\n",
                           ppid, username->data, domain->data);
                  exit (0);
                }
            }
        }
    }

  //Now we should have the following
  //
  // valid = 1 if the user exists or if it's not local
  // valid = 0 if the user is local and does not exist
  //
  // user_passwd containing the data for the real user (alias already followed)
  // username and domain with the right data

  if (valid == 0)               //Not valid, not relaying, we don't like it
    {
      wrongrcptcount = wrongrcptcount + 1;
      printf ("SWRONGRCPTCOUNT=%d\n", wrongrcptcount);
    }

  // Over $CHKUSER_WRONGRCPTLIMIT we don't care if it's valid or not, by policy
  if (wrongrcptcount > chkuser_wrongrcptlimit)
    {
      printf
        ("E571 sorry, you are violating our security policies (#5.7.1 - chkuser2)\n");
      fprintf (stderr,
               "rcptchecks: pid %d - 571 sorry, you are violating our security policies (%d of %d - %s@%s)\n",
               ppid, wrongrcptcount, chkuser_wrongrcptlimit, username->data,
               domain->data);
      exit (0);
    }

  //Finally, if it's not valid, but we are within limits, plain doesn't exist
  if (valid == 0)
    {
      printf
        ("E511 sorry, no mailbox here by that name (#5.1.1 - chkuser2)\n");
      fprintf (stderr,
               "rcptchecks: pid %d - 511 sorry, no mailbox here by that name (%s@%s)\n",
               ppid, username->data, domain->data);
      exit (0);
    }
  fprintf (stderr,
           "rcptchecks: pid %d - Accepted recipient (%s@%s number %d of %d )\n",
           ppid, username->data, domain->data, smtprcptcount,
           chkuser_rcptlimit);
  exit (0);
}
